"""Main interface to the duotecno bus."""

from __future__ import annotations
import asyncio
import logging
import time
from typing import Final
from collections import deque
from duotecno.exceptions import LoadFailure, InvalidPassword
from duotecno.protocol import (
    Packet,
    EV_CLIENTCONNECTSET_3,
    EV_NODEDATABASEINFO_0,
    EV_NODEDATABASEINFO_1,
    EV_HEARTBEATSTATUS_1,
)
from duotecno.node import Node
from duotecno.unit import BaseUnit

PW_TIMEOUT: Final = 5
LOAD_NODE_TIMEOUT: Final = 60
LOAD_UNIT_TIMEOUT: Final = 120
HB_TIMEOUT: Final = 20
HB_BUSEMPTY: Final = 60
MAX_INFLIGHT: Final = 5
STATUS_RETRANSMIT: Final = 2
RECONNECT_TIMEOUT: Final = 180


class PyDuotecno:
    """Class that will will do the bus management.

    - send packets
    - receive packets
    - open and close the connection
    """

    writer: asyncio.StreamWriter | None = None
    reader: asyncio.StreamReader | None = None
    readerTask: asyncio.Task[None]
    hbTask: asyncio.Task[None]
    workTask: asyncio.Task[None]
    writerTask: asyncio.Task[None]
    receiveQueue: asyncio.PriorityQueue
    sendSema: asyncio.Semaphore
    sendQueue: asyncio.PriorityQueue
    connectionOK: asyncio.Event
    heartbeatReceived: asyncio.Event
    nextHeartbeat: int
    packetToWaitFor: str | None = None
    packetWaiter: asyncio.Event
    nodes: dict[int, Node] = {}
    host: str
    port: int
    password: str
    numNodes: int = 0

    def get_units(self, unit_type: list[str] | str) -> list[BaseUnit]:
        res = []
        for node in self.nodes.values():
            for unit in node.get_unit_by_type(unit_type):
                res.append(unit)
        return res

    async def enableAllUnits(self) -> None:
        self._log.debug("Enable all Units on all nodes")
        for node in self.nodes.values():
            await node.enable()

    async def disableAllUnits(self) -> None:
        self._log.debug("Disable all Units on all nodes")
        for node in self.nodes.values():
            await node.disable()

    async def disconnect(self) -> None:
        self._log.debug("Disconnecting")
        self.connectionOK.clear()

        if self.readerTask.cancel():
            try:
                await self.readerTask
            except asyncio.CancelledError:
                pass
        if self.hbTask.cancel():
            try:
                await self.hbTask
            except asyncio.CancelledError:
                pass
        if self.workTask.cancel():
            try:
                await self.workTask
            except asyncio.CancelledError:
                pass
        if self.writerTask.cancel():
            try:
                await self.writerTask
            except asyncio.CancelledError:
                pass
        if self.writer:
            self.writer.close()
        self._log.debug("Disconnecting Finished")

    async def connect(
        self, host: str, port: int, password: str, testOnly: bool = False
    ) -> None:
        """Initialize the connection."""
        self.host = host
        self.port = port
        self.password = password
        await self._do_connect(testOnly)

    async def _reconnect(self):
        await self.disconnect()
        await self.disableAllUnits()
        await asyncio.sleep(RECONNECT_TIMEOUT)
        await self.continuously_check_connection()

    async def _do_connect(self, testOnly: bool = False, skipLoad: bool = False) -> None:
        if not skipLoad:
            self.nodes = {}
        self._log = logging.getLogger("pyduotecno")
        # Try to connect
        self._log.debug("Try to connect")
        try:
            self.reader, self.writer = await asyncio.open_connection(
                self.host, self.port
            )
        except (ConnectionError, TimeoutError):
            raise
        # events
        self.connectionOK = asyncio.Event()
        self.heartbeatReceived = asyncio.Event()
        self.packetWaiter = asyncio.Event()
        # at this point the connection should be ok
        self._log.debug("Connection established")
        self.connectionOK.set()
        self.heartbeatReceived.clear()
        # start the bus reading task
        self.receiveQueue = asyncio.PriorityQueue()
        self.sendQueue = asyncio.PriorityQueue()
        self.sendSema = asyncio.Semaphore(MAX_INFLIGHT)
        self.readerTask = asyncio.Task(self._readTask())
        self.writerTask = asyncio.Task(self._writeTask())
        self.workTask = asyncio.Task(self._handleTask())
        # send login info
        passw = [str(ord(i)) for i in self.password]
        await self.write(f"[214,3,{len(passw)},{','.join(passw)}]")
        # wait for the login to be ok
        try:
            await asyncio.wait_for(self.waitForPacket("67,3,1"), timeout=PW_TIMEOUT)
        except TimeoutError:
            await self.disconnect()
            raise InvalidPassword()
        # if we are not testing the connection, start scanning
        if testOnly:
            return
        # do we need to reload the modules?
        if not skipLoad:
            await self.write("[209,5]")
            await self.write("[209,0]")
            try:
                await asyncio.wait_for(self._loadTaskNodes(), timeout=LOAD_NODE_TIMEOUT)
                self._log.info("Nodes discoverd")
                for n in self.nodes.values():
                    await n.load()
                    await asyncio.sleep(0.1)
                await asyncio.wait_for(self._loadTaskUnits(), timeout=LOAD_UNIT_TIMEOUT)
                self._log.info("Units discoverd")
            except TimeoutError:
                await self.disconnect()
                raise LoadFailure()
        # in case of skipload we do want to request the status again
        self._log.info("Requesting unit status")
        for node in self.nodes.values():
            for unit in node.get_units():
                self._log.debug(f"Unit: {unit}")
                await unit.requestStatus()
                await asyncio.sleep(0.1)
        self.hbTask = asyncio.Task(self.heartbeatTask())
        await self.enableAllUnits()

    async def write(self, msg: str) -> None:
        """Send a message."""
        if not self.writer:
            return
        if self.writer.transport.is_closing():
            await self._reconnect()
            return
        self._log.debug(f"TX: {msg}")
        await self.sendQueue.put(msg)
        return

    async def _writeTask(self) -> None:
        while True:
            try:
                msg = await self.sendQueue.get()
                await self.sendSema.acquire()
                msg = f"{msg}{chr(10)}"
                self.writer.write(msg.encode())
                await self.writer.drain()
                await asyncio.sleep(0.1)
            except ConnectionError:
                await self.reconnect()
                return

    async def _loadTaskNodes(self) -> None:
        while len(self.nodes) < 1:
            await asyncio.sleep(3)
        while True:
            if len(self.nodes) == self.numNodes:
                return
            await asyncio.sleep(1)

    async def _loadTaskUnits(self) -> None:
        while True:
            c = 0
            for n in self.nodes.values():
                if n.isLoaded.is_set():
                    c += 1
                if c == len(self.nodes):
                    return
            await asyncio.sleep(1)

    async def check_tcp_connection(self, timeout=3) -> bool:
        """Check if a TCP connection can be established to the given host and port."""
        self._log.debug("Checking connection...")
        conn = asyncio.open_connection(self.host, self.port)
        try:
            reader, writer = await asyncio.wait_for(conn, timeout=timeout)
            writer.close()
            await writer.wait_closed()
            return True
        except (asyncio.TimeoutError, OSError):
            # Could not connect within the timeout period or there was a network error
            return False

    async def continuously_check_connection(self) -> None:
        """Continuously check for connection restoration and reconnect."""
        self._log.info("Waiting for connection...")
        while True:
            connection_restored = await self.check_tcp_connection()
            if connection_restored:
                self._log.info("Connection to host restored, reconnecting.")
                await self._do_connect(skipLoad=True)
                break
            else:
                self._log.debug("Connection to host not yet restored, retrying...")
                await asyncio.sleep(5)

    async def heartbeatTask(self) -> None:
        await asyncio.sleep(30)
        self._log.info("Starting HB task")
        while True:
            # wait until the timer expire of 5 seconds
            while self.nextHeartbeat > int(time.time()):
                self._log.debug(
                    f"Waiting until {self.nextHeartbeat} to send a HB ({time.time()})"
                )
                await asyncio.sleep(5)
            # send the heartbeat
            self.heartbeatReceived.clear()
            try:
                self._log.debug("Sending heartbeat message")
                await self.write("[215,1]")
                await asyncio.wait_for(
                    self.heartbeatReceived.wait(), timeout=HB_TIMEOUT
                )
                self._log.debug("Received heartbeat message")
            except TimeoutError:
                self._log.warning("Timeout on heartbeat, reconnecting")
                await self._reconnect()
                break
            except asyncio.exceptions.CancelledError:
                break
        self._log.info("Stopping HB task")

    async def _readTask(self) -> None:
        """Reader task."""
        while self.connectionOK.is_set() and self.reader:
            try:
                tmp2 = await self.reader.readline()
            except ConnectionError:
                await self._reconnect()
                return
            if tmp2 == "":
                return
            tmp3 = tmp2.decode()
            tmp = tmp3.rstrip()
            # self._log.debug(f'Raw Receive: "{tmp}"')
            if not tmp.startswith("["):
                tmp = tmp.lstrip("[")
            tmp = tmp.replace("\x00", "")
            # self._log.debug(f'Receive: "{tmp}"')
            tmp = tmp[1:-1]
            if tmp == "":
                return
            self._log.debug(f'RX: "{tmp}"')
            self.nextHeartbeat = int(time.time()) + HB_BUSEMPTY
            self.sendSema.release()
            if await self._comparePacket(tmp):
                p = tmp.split(",")
                try:
                    pc = Packet(int(p[0]), int(p[1]), deque([int(_i) for _i in p[2:]]))
                    await self.receiveQueue.put(pc)
                # self._log.debug(f'QUEUE: "{pc}" {self.receiveQueue.qsize()}')
                except Exception as e:
                    self._log.error(e)
                    self._log.error(tmp)
            await asyncio.sleep(0.1)

    async def _comparePacket(self, rpck: str) -> bool:
        if not self.packetToWaitFor:
            return True
        # self._log.debug(f"COMPARE received packet {rpck} with {self.packetToWaitFor}")
        if rpck.startswith(self.packetToWaitFor):
            self.packetWaiter.set()
            return True
        return False

    async def waitForPacket(self, pstr: str) -> None:
        """Wait for a certain packet.

        will clear an event and then wait for the event to be set again
        """
        self.packetWaiter.clear()
        self.packetToWaitFor = pstr
        await self.packetWaiter.wait()
        self.packetToWaitFor = None

    async def _handleTask(self) -> None:
        """handler task."""
        while self.connectionOK.is_set() and self.receiveQueue:
            try:
                pc = await self.receiveQueue.get()
                self._log.debug(f"WX: {pc}")
                await self._handlePacket(pc)
            except Exception as e:
                self._log.error(e)
            await asyncio.sleep(0.1)

    async def _handlePacket(self, packet: Packet) -> None:
        if packet.cls is None:
            self._log.debug(f"Ignoring packet: {packet}")
            return
        if isinstance(packet.cls, EV_CLIENTCONNECTSET_3):
            return
        if isinstance(packet.cls, EV_HEARTBEATSTATUS_1):
            self.heartbeatReceived.set()
            return
        if isinstance(packet.cls, EV_NODEDATABASEINFO_0):
            self.numNodes = packet.cls.numNode
            for i in range(packet.cls.numNode):
                await self.write(f"[209,1,{i}]")
                await self.waitForPacket(f"64,1,{i}")
            return
        if isinstance(packet.cls, EV_NODEDATABASEINFO_1):
            if packet.cls.address not in self.nodes:
                self.nodes[packet.cls.address] = Node(
                    name=packet.cls.nodeName,
                    address=packet.cls.address,
                    index=packet.cls.index,
                    nodeType=packet.cls.nodeType,
                    numUnits=packet.cls.numUnits,
                    writer=self.write,
                    pwaiter=self.waitForPacket,
                )
                # await self.nodes[packet.cls.address].load()
            return
        if hasattr(packet.cls, "address") and packet.cls.address in self.nodes:
            await self.nodes[packet.cls.address].handlePacket(packet.cls)
            return
        self._log.debug(f"Ignoring packet: {packet}")



from __future__ import annotations
from typing import final, Deque
from enum import Enum, unique
from dataclasses import dataclass, field
import collections
import sys
import json


@unique
class MsgType(Enum):
    EV_UNITCONTROLSTATUS = 4
    EV_UNITDIMSTATUS = 5
    EV_UNITSWITCHSTATUS = 6
    EV_UNITSENSSTATUS = 7
    EV_MESSAGEERROR = 17
    EV_NODERESET = 18
    EV_UNITAUDIOSTATUS = 23
    EV_UNITDUOSWITCHSTATUS = 38
    EV_UNITAVMATRIXSTATUS = 54
    EV_UNITDEFAULTSTATUS = 48
    EV_NODEDATABASEINFO = 64
    EV_APPLICATIONTASKSTATUS = 66
    EV_CLIENTCONNECTSET = 67
    EV_UNITMACROCOMMAND = 69
    EV_UNITAUDIOEXTSTATUS = 70
    EV_TIMEDATESTATUS = 71
    EV_HEARTBEATSTATUS = 72
    EV_SCHEDULESTATUS = 73
    EV_NODEMANAGEMENTINFO = 74
    EV_ACCESSLEVELSET = 75
    EV_VIDEOPHONESTATUS = 76
    EV_REGISTERMAP = 77
    FC_UNITDIMREQUESTSTATUS = 131
    FC_UNITSENSSET = 136
    FC_UNITREQUESTSENSSTATUS = 137
    FC_NODERESETSET = 155
    FC_UNITAUDIOBASICSET = 159
    FC_UNITDIMSET = 162
    FC_UNITSWITCHSET = 163
    FC_UNITCONTROLSET = 168
    FC_TIMEDATE = 170
    FC_UNITIRTXSET = 173
    FC_UNITDUOSWITCHSET = 182
    FC_CHECKIRRXCODE = 192
    FC_UNITVIDEOMUXSET = 193
    FC_UNITAVMATRIXSET = 202
    FC_UNITALARMSET = 204
    FC_UNITAUDIOEXTSET = 208
    FC_NODEDATABASEREQUESTSTATUS = 209
    FC_APPLICATIONTASKSET = 212
    FC_REQUESTAPPLICATIONTASKSTATUS = 213
    FC_CLIENTCONNECTSET = 214
    FC_HEARTBEATREQUESTSTATUS = 215
    FC_REQUESTTIMEDATE = 216
    FC_SCHEDULESET = 217
    FC_REQUESTSCHEDULE = 218
    FC_NODEMANAGEMENTSET = 219
    FC_REQUESTNODEMANAGEMENT = 220
    FC_NODEDATABASESET = 221
    FC_ACCESSLEVELSET = 222
    FC_VIDEOPHONESET = 223
    FC_REGISTERMAP = 224
    UNKNOWN = 255

    @classmethod
    def _missing_(cls, value: object) -> MsgType:
        return cls.UNKNOWN


@dataclass
class Packet:
    """Basic structure for a packet."""

    cmdName: str = field(init=False)
    cmdCode: int = field(repr=False)
    method: int
    data: Deque[int]
    cls: BaseMessage | None = field(init=False)

    def __lt__(self, other: Packet) -> bool:
        if isinstance(other.cls, EV_HEARTBEATSTATUS_1):
            return True
        return False

    def __post_init__(self) -> None:
        """fill in the command name, make the subsclass."""
        try:
            self.cmdName = MsgType(self.cmdCode).name
        except ValueError:
            self.cmdName = "UNKNOWN"
        self.data = collections.deque(self.data)
        tmp = getattr(sys.modules[__name__], f"{self.cmdName}_{self.method}", None)
        if tmp:
            self.cls = tmp(self.data)
            # self.data should be empty once the message consumed it
            if len(self.data) != 0:
                print(f"ERROR!!! Not all data consumed: {self}")
        else:
            self.cls = None


def calc_value(msb: int, lsb: int) -> int:
    return (256 * msb) + lsb


class BaseMessage:
    def __init__(self, data: Deque[int]) -> None:
        pass

    def to_json(self) -> str:
        return json.dumps(self.to_json_basic())

    def to_json_basic(self) -> dict[str, str]:
        """
        Create JSON structure with generic attributes
        """
        me = {}
        me["name"] = str(self.__class__.__name__)
        me.update(self.__dict__.copy())
        for key in me.copy():
            if key == "name":
                continue
            if isinstance(me[key], str):
                continue
            if callable(getattr(self, key)) or key.startswith("__"):
                del me[key]
            if isinstance(me[key], Enum):
                me[key] = me[key].name
            if isinstance(me[key], (bytes, bytearray)):
                me[key] = str(me[key], "utf-8")
        return me

    def __repr__(self) -> str:
        return self.to_json()


class BaseNodeUnitMessage(BaseMessage):
    address: int
    unit: int

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        self.address = data.popleft()
        self.unit = data.popleft()


class BaseNodeUnitTypeMessage(BaseNodeUnitMessage):
    unitType: int

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        self.unitType = data.popleft()


class EV_HEARTBEATSTATUS_1(BaseMessage):
    pass


class EV_CLIENTCONNECTSET_3(BaseMessage):
    loginOk: bool

    def __init__(self, data: Deque[int]) -> None:
        self.loginOK = data.popleft()


@unique
class DbState(Enum):
    Empty = 0
    Busy = 1
    Ready = 2


class EV_NODEDATABASEINFO_5(BaseMessage):
    state: DbState

    def __init__(self, data: Deque[int]) -> None:
        self.state = data.popleft()


class EV_NODEDATABASEINFO_0(BaseMessage):
    numNode: int

    def __init__(self, data: Deque[int]) -> None:
        self.numNode = data.popleft()


class EV_UNITMACROCOMMAND_0(BaseNodeUnitMessage):
    event: int
    state: int
    code1: int
    code2: int

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        self.event = data.popleft()
        self.state = data.popleft()
        self.code1 = data.popleft()
        self.code2 = data.popleft()


@unique
class NodeType(Enum):
    Standard = 1
    Gateway = 4
    Modem = 8
    Gui = 32
    UNKNOWN = 255

    @classmethod
    def _missing_(cls, value: object) -> NodeType:
        return cls.UNKNOWN


class EV_NODEDATABASEINFO_1(BaseMessage):
    index: int
    address: int
    nodeName: str
    numUnits: int
    nodeType: NodeType
    nodeTypeName: str
    flags: int

    def __init__(self, data: Deque[int]) -> None:
        self.index = data.popleft()
        self.address = data.popleft()
        # next 4 are no needed
        [data.popleft() for _i in range(4)]
        self.nodeName = "".join([chr(data.popleft()) for _i in range(data.popleft())])
        self.numUnits = data.popleft()
        self.nodeType = NodeType(data.popleft())
        self.nodeTypeName = NodeType(self.nodeType).name
        self.flags = data.popleft()


@unique
class UnitType(Enum):
    DIM = 1
    SWITCH = 2
    CONTROL = 3
    SENS = 4
    AUDIO_EXT = 5
    VIRTUAL = 7
    DUOSWITCH = 8
    AUDIO_BASIC = 10
    AVMATRIC = 11
    IRTX = 12
    VIDEOMUX = 14
    UNKNOWN = 255

    @classmethod
    def _missing_(cls, value: object) -> UnitType:
        return cls.UNKNOWN


class EV_NODEDATABASEINFO_2(BaseMessage):
    address: int
    unit: int
    laddress: int
    lunit: int
    unitName: str
    unitType: int
    unitTypeName: str
    unitFlags: int

    def __init__(self, data: Deque[int]) -> None:
        self.address = data.popleft()
        self.unit = data.popleft()
        self.laddress = data.popleft()
        self.lunit = data.popleft()
        self.unitName = "".join([chr(data.popleft()) for _i in range(data.popleft())])
        self.unitType = data.popleft()
        self.unitTypeName = UnitType(self.unitType).name
        self.unitFlags = data.popleft()


@final
@unique
class SwitchStatus(Enum):
    OFF = 0
    ON = 1
    PIRTIMED = 2


class EV_UNITSWITCHSTATUS_0(BaseNodeUnitTypeMessage):
    state: int
    stateName: str

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        # config, reserved
        data.popleft()
        self.state = data.popleft()
        self.stateName = SwitchStatus(self.state).name


class EV_UNITDIMSTATUS_0(BaseNodeUnitTypeMessage):
    state: int
    stateName: str
    dimValue: int

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        # config, reserved
        data.popleft()
        self.state = data.popleft()
        self.stateName = SwitchStatus(self.state).name
        self.dimValue = data.popleft()


@final
@unique
class DuoswitchStatus(Enum):
    IDLE = 0
    IDLE_DOWN = 1
    IDLE_UP = 2
    BUSY_DOWN = 3
    BUSY_UP = 4


class EV_UNITDUOSWITCHSTATUS_0(BaseNodeUnitTypeMessage):
    state: int
    stateName: str

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        # config, reserved
        data.popleft()
        self.state = data.popleft()
        self.stateName = DuoswitchStatus(self.state).name


@final
@unique
class SensType(Enum):
    TEMPERATURE = 0
    PH = 1
    LUX = 2
    AMPERE = 3


@final
@unique
class SensControl(Enum):
    OFF = 0
    ON = 1


@final
@unique
class SensState(Enum):
    IDLE = 0
    HEATING = 1
    COOLING = 2


@final
@unique
class SensPreset(Enum):
    SUN = 0
    HALF_SUN = 1
    MOON = 2
    HALF_MOON = 3


@final
@unique
class SensWorkingmode(Enum):
    AUTO = 0
    HEATING = 1
    COOLING = 2
    DRY = 3
    FAN = 4
    UNKNOWN = 255


@final
@unique
class SensFanspeed(Enum):
    SPEED1 = 0
    SPEED2 = 1
    SPEED3 = 2
    SPEED4 = 3
    SPEED5 = 4
    AUTO = 255


def sens_calc_value(msb: int, lsb: int) -> float:
    val = calc_value(msb, lsb)
    if (val & (1 << 15)) != 0:
        val = val - (1 << 16)
    return val / 10


class EV_UNITSENSSTATUS_0(BaseNodeUnitTypeMessage):
    config: int
    configName: str
    controlState: int
    controlStateName: str
    state: int
    stateName: str
    preset: int
    presetName: str
    value: float
    sun: float
    halfsun: float
    moon: float
    halfmoon: float

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        self.config = data.popleft()
        self.configName = SensType(self.config).name
        self.controlState = data.popleft()
        self.controlStateName = SensControl(self.controlState).name
        self.state = data.popleft()
        self.stateName = SensState(self.state).name
        self.preset = data.popleft()
        self.presetName = SensPreset(self.preset).name
        self.value = sens_calc_value(data.popleft(), data.popleft())
        self.sun = sens_calc_value(data.popleft(), data.popleft())
        self.halfsun = sens_calc_value(data.popleft(), data.popleft())
        self.moon = sens_calc_value(data.popleft(), data.popleft())
        self.halfmoon = sens_calc_value(data.popleft(), data.popleft())


class EV_UNITSENSSTATUS_1(EV_UNITSENSSTATUS_0):
    offset: float
    swing: float
    workingMode: int
    workingModeName: str
    fanSpeed: int
    fanSpeedName: str
    swingMode: int
    swingModeName: str

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        self.offset = sens_calc_value(data.popleft(), data.popleft())
        self.swing = sens_calc_value(data.popleft(), data.popleft())
        self.workingMode = data.popleft()
        self.workingModeName = SensWorkingmode(self.workingMode).name
        self.fanSpeed = data.popleft()
        self.fanSpeedName = SensFanspeed(self.fanSpeed).name
        self.swingMode = data.popleft()
        self.swingModeName = SensControl(self.swingMode).name


@final
@unique
class ControLStatus(Enum):
    OFF = 0
    ON = 1


class EV_UNITCONTROLSTATUS_0(BaseNodeUnitTypeMessage):
    status: int
    statusName: str

    def __init__(self, data: Deque[int]) -> None:
        super().__init__(data)
        # config ignore
        data.popleft()
        self.status = data.popleft()
        self.statusName = ControLStatus(self.status).name



"""Basic example to read the bus."""
import argparse
import asyncio
import logging
import sys
from duotecno.controller import PyDuotecno


async def test(host, port, passw):
    """Basic aio call."""
    tmp = PyDuotecno()
    await tmp.connect(host, port, passw)
    await asyncio.sleep(6000000000)


parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--host", help="The host to connect to")
parser.add_argument("--port", help="The port to connect to")
parser.add_argument("--password", help="The password")
args = parser.parse_args()

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logging.getLogger("asyncio").setLevel(logging.DEBUG)

asyncio.run(test(args.host, args.port, args.password))



from __future__ import annotations
from typing import Callable, Awaitable
import asyncio
import logging

from duotecno.protocol import NodeType, EV_NODEDATABASEINFO_2, BaseMessage
from duotecno.unit import (
    BaseUnit,
    SwitchUnit,
    SensUnit,
    DimUnit,
    DuoswitchUnit,
    VirtualUnit,
    ControlUnit,
)


class Node:
    name: str
    index: int
    nodeType: NodeType
    address: int
    numUnits: int
    units: dict[int, BaseUnit]
    isloaded: asyncio.Event

    def __init__(
        self,
        name: str,
        address: int,
        index: int,
        nodeType: NodeType,
        numUnits: int,
        writer: Callable[[str], Awaitable[None]],
        pwaiter: Callable[[str], Awaitable[None]],
    ) -> None:
        self._log = logging.getLogger("pyduotecno-node")
        self.name = name
        self.address = address
        self.index = index
        self.numUnits = numUnits
        self.nodeType = nodeType
        self.writer = writer
        self.pwaiter = pwaiter
        self.isLoaded = asyncio.Event()
        self.isLoaded.clear()
        self.units = {}
        self._log.info(f"New node found: {self.name}")

    async def enable(self) -> None:
        for unit in self.units.values():
            await unit.enable()

    async def disable(self) -> None:
        for unit in self.units.values():
            await unit.disable()

    def get_name(self) -> str:
        return self.name

    def get_address(self) -> int:
        return self.address

    def __repr__(self) -> str:
        items = []
        for k, v in self.__dict__.items():
            if k not in ["_log", "writer"]:
                items.append(f"{k} = {v!r}")
        return "{}[{}]".format(type(self), ", ".join(items))

    def get_units(self) -> list[BaseUnit]:
        res = []
        for unit in self.units.values():
            res.append(unit)
        return res

    def get_unit_by_type(self, unit_type: list[str] | str) -> list[BaseUnit]:
        if isinstance(unit_type, str):
            unit_type = [unit_type]
        res = []
        for unit in self.units.values():
            for unitT in unit_type:
                if str(type(unit)) == f"<class 'duotecno.unit.{unitT}'>":
                    res.append(unit)
        return res

    async def load(self) -> None:
        self._log.debug(f"Node {self.name}: Requesting units")
        for i in range(self.numUnits):
            await self.writer(f"[209,2,{self.address},{i}]")
            await self.pwaiter(f"64,2,{self.address},{i}")

    async def handlePacket(self, packet: BaseMessage) -> None:
        if isinstance(packet, EV_NODEDATABASEINFO_2):
            if packet.unit not in self.units:
                u = BaseUnit
                if packet.unitTypeName == "SWITCH":
                    u = SwitchUnit
                elif packet.unitTypeName == "SENS":
                    u = SensUnit
                elif packet.unitTypeName == "DIM":
                    u = DimUnit
                elif packet.unitTypeName == "DUOSWITCH":
                    u = DuoswitchUnit
                elif packet.unitTypeName == "VIRTUAL":
                    u = VirtualUnit
                elif packet.unitTypeName == "CONTROL":
                    u = ControlUnit
                else:
                    self._log.warning(f"Unhandled unitType: {packet.unitTypeName}")
                self.units[packet.unit] = u(
                    self, name=packet.unitName, unit=packet.unit, writer=self.writer
                )
            if len(self.units) == self.numUnits:
                self.isLoaded.set()
            return
        if hasattr(packet, "unit") and packet.unit in self.units:
            await self.units[packet.unit].handlePacket(packet)
            return
